import ast
import dataclasses
import inspect
import multiprocessing
import os.path
from concurrent.futures import ProcessPoolExecutor
from contextlib import nullcontext
from dataclasses import dataclass
from functools import cached_property, partial
from typing import Callable, Sequence, Set, Optional, Dict, FrozenSet, List

import h5py
import numpy as np

from .. import unstack, stack, batch
from ..backend import ML_LOGGER
from ..dataclasses._dep import MemberVariableAnalyzer
from ..dataclasses._tensor_cache import H5Source, write_to_h5
from ..math import DimFilter, shape, Shape, EMPTY_SHAPE
from ..math._magic_ops import all_attributes
from ..math._tensors import disassemble_tree, assemble_tree


def parallel_compute(instance, properties: Sequence, parallel_dims=batch,
                     max_workers=multiprocessing.cpu_count(), memory_limit: Optional[float] = None, cache_dir: str = None, keep_intermediate=False):
    """
    Compute the values of properties decorated with `@cached_property` or `@parallel_property` of a dataclass instance in parallel.

    **Multiple stages via `requires=...`**
    If `@parallel_property` are computed whose `requires` overlaps with `parallel_dims`, a separate computation stage is set up to compute these properties with fewer parallel workers.
    In the presence of different `requires`, the computation is split into different stages in accordance with the property dependency graph.
    Properties that cannot be parallelized (because it requires all `parallel_dims`) are computed on the host process.

    **Caching tensors to disk**
    When `memory_limit` and `cache_dir` are set, the evaluation will try to adhere to the given memory limits by moving tensors out of memory onto disk.
    This is only applied to the outputs of `@cached_property` and `@parallel_property` calls, not to intermediate values used in their computation.
    The per-process memory limit is calculated per stage, dividing the total memory by the active worker count.
    Cached tensors behave like regular tensors and are temporarily loaded back into memory when accessing their values or using them in a computation.
    When parallelizing, the full result is assembled by stacking multiple disk-backed tensors from different files created by different processes.
    These composite tensors will reference multiple binary files and can be pickled/unpickled safely without loading the data into memory.
    This enables passing large data references to different processes or saving the structure to a file without the data content.

    See Also:
        `cached_property`, `parallel_property`, `get_cache_files()`, `on_load_into_memory()`.

    Warnings:
        `parallel_compute` breaks automatic differentiation.

    Args:
        instance: Dataclass instance for which to compute the values of `@cached_property` or `@parallel_property` fields.
        properties: References to the unbound properties. These must be `cached_property` or `parallel_property`.
        parallel_dims: Dimensions to parallelize over.
        max_workers: Number of processes to spawn.
        memory_limit: Limit to the total memory consumption from `Tensor` instances on property outputs.
        cache_dir: Directory path to store cached tensors in if `memory_limit` is set.
        keep_intermediate: Whether the outputs of cached properties required to compute `properties` but not contained in `properties` should be kept in memory.
            If `False`, these values will not be cached on `instance` after this call.
    """
    if memory_limit is not None:
        assert cache_dir is not None, "cache_dir must be specified if memory_limit is set"
    dims = shape(instance).only(parallel_dims)
    # --- Build graph of relevant properties ---
    cls = type(instance)
    nodes: Dict[str, PGraphNode] = {}
    output_user = PGraphNode('<output>', None, None, None, False, [], 999999)
    for p in properties:
        recursive_add_node(cls, property_name(p), p, dims, nodes).users.append(output_user)
    for node in nodes.values():
        if node.name in instance.__dict__:
            node.done = True
            node.dependencies = []
    stages = build_stages(nodes)
    ML_LOGGER.debug(f"Assembled {len(stages)} stages containing {sum(len(ns) for ns in stages)} properties for parallel computation.")
    # --- Execute stages ---
    any_parallel = any(dims - tuple(stage_nodes[0].requires) for stage_nodes in stages)
    with ProcessPoolExecutor(max_workers=max_workers) if any_parallel else nullcontext() as pool:
        for stage_idx, stage_nodes in enumerate(stages):
            parallel_dims = dims - tuple(stage_nodes[0].requires)
            if parallel_dims:
                ML_LOGGER.debug(f"Parallel | {parallel_dims} | {[n.name for n in stage_nodes]}")
                property_names = [n.name for n in stage_nodes if n.is_used_later]
                required_caches = set.union(*[n.all_dep_names for n in stage_nodes])
                # --- Submit to pool ---
                instances = unstack(instance, parallel_dims)
                caches = [unstack(instance.__dict__[c], parallel_dims, expand=True) for c in required_caches]
                caches = [{c: caches[j][i] for j, c in enumerate(required_caches)} for i in range(len(instances))] if caches else [{}] * len(instances)
                keep_intermediate or delete_intermediate_caches(instance, stages, stage_idx)
                inst_props = [property_names] * len(instances)
                mem_per_item = memory_limit / min(max_workers, len(instances)) if memory_limit is not None else None
                cache_dir is not None and os.makedirs(cache_dir, exist_ok=True)
                cache_file_suggestions = [os.path.join(cache_dir, f"s{stage_idx}_i{i}") for i in range(len(instances))] if cache_dir is not None else [None] * len(instances)
                results = list(pool.map(_evaluate_properties, instances, inst_props, caches, [mem_per_item]*len(instances), cache_file_suggestions))
                for name, *outputs in zip(property_names, *results):
                    output = stack(outputs, parallel_dims)
                    instance.__dict__[name] = output
            else:  # No parallelization in this stage
                ML_LOGGER.debug(f"Host | {[n.name for n in stage_nodes]}")
                for n in stage_nodes:
                    get_property_value(instance, n.name)
                keep_intermediate or delete_intermediate_caches(instance, stages, stage_idx)


def delete_intermediate_caches(instance, stages: list, stage_idx: int):
    for node in sum(stages[:stage_idx+1], []):
        if not node.has_users_after(stage_idx):
            ML_LOGGER.debug(f"Host | Removing cache of {node.name} as it's not needed after stage {stage_idx}")
            del instance.__dict__[node.name]


def _evaluate_properties(instance, properties, cache: dict, mem_limit, cache_file_base: Optional[str]):  # called on workers
    instance.__dict__.update(cache)
    # print(f"Evaluating {properties} on {type(instance).__name__} with values {instance.__dict__}")
    values = [get_property_value(instance, p) for p in properties]
    if mem_limit is not None:
        tree, tensors = disassemble_tree(values, False, attr_type=all_attributes)
        sizes = [sum([t.backend.sizeof(nat) for nat in t._natives()]) for t in tensors]
        if sum(sizes) > mem_limit:
            sorted_pairs = sorted(zip(sizes, range(len(tensors))))
            sizes, indices = zip(*sorted_pairs)
            total_sizes = np.cumsum(sizes)
            cutoff = np.argmax(total_sizes > mem_limit)
            file_counter = 0
            while True:
                cache_file = f"{cache_file_base}_{file_counter}.h5"
                try:
                    with h5py.File(cache_file, 'x') as f:  # fail if exists
                        ML_LOGGER.debug(f"Result is too large for memory limit. Tensor sizes (bytes): {sizes[::-1]}. Caching {len(tensors) - cutoff} tensors to {cache_file}.")
                        h5_ref = H5Source(cache_file)
                        for i in range(len(tensors) - cutoff - 1, len(tensors)):
                            tensors[i] = write_to_h5(tensors[i], f't{i}', f, h5_ref)
                        break
                except FileExistsError:
                    file_counter += 1
            values = assemble_tree(tree, tensors, attr_type=all_attributes)
    return values


def get_dependencies(cls: type, cls_property) -> Set[str]:
    if hasattr(cls, '__phiml_parallel_dep__') and cls_property in cls.__phiml_parallel_dep__:
        return cls.__phiml_parallel_dep__[cls_property]
    if isinstance(cls_property, cached_property):
        method = cls_property.func
    elif isinstance(cls_property, property):
        method = cls_property.fget
    else:
        assert callable(cls_property) and hasattr(cls_property, '__qualname__'), f"Dependency resolver failed on {cls_property} of {cls.__name__}"
        method = cls_property
    ML_LOGGER.debug(f"Analyzing dependencies of {method.__qualname__}")
    source_code = inspect.getsource(method)
    indent0 = len(source_code) - len(source_code.lstrip())
    source_code_top = "\n".join([line[indent0:] for line in source_code.split("\n")])
    tree = ast.parse(source_code_top)
    analyzer = MemberVariableAnalyzer()
    analyzer.visit(tree)
    direct_deps = analyzer.member_vars
    fields = set([f.name for f in dataclasses.fields(cls)])
    all_caches = set([m for m in dir(cls) if isinstance(getattr(cls, m), cached_property)])
    prop_deps = all_caches & direct_deps
    # --- Indirect dependencies via method/property calls ---
    for method_dep in direct_deps - fields - prop_deps:
        if not hasattr(cls, method_dep):
            continue  # may be a dynamic dim, such as vector
        dep = getattr(cls, method_dep)
        if callable(dep) or isinstance(dep, property):
            prop_deps.update(get_dependencies(cls, dep))
    # --- Cache result ---
    if not hasattr(cls, '__phiml_parallel_dep__'):
        cls.__phiml_parallel_dep__ = {cls_property: prop_deps}
    else:
        cls.__phiml_parallel_dep__[cls_property] = prop_deps
    return prop_deps


@dataclass
class PGraphNode:
    name: str
    prop: property
    requires: FrozenSet[str]
    dependencies: Sequence['PGraphNode'] = None
    done: bool = False
    users: List['PGraphNode'] = dataclasses.field(default_factory=lambda: [])
    stage: int = None

    def __repr__(self):
        return f"{self.name}{f'@{self.stage}' if self.done else ' (pending)'}->{[n.name for n in self.dependencies]}<-{len(self.users)}"

    @property
    def can_run_now(self):
        return not self.done and all(dep.done or (dep.can_run_now and dep.requires == self.requires) for dep in self.dependencies)

    @property
    def is_used_later(self):  # requires caching
        return any(u.stage != self.stage for u in self.users)  # output has stage -1

    @property
    def all_dep_names(self) -> Set[str]:
        return set.union(*[{dep.name, *dep.all_dep_names} for dep in self.dependencies]) if self.dependencies else set()

    def has_users_after(self, stage: int):
        return any(u.stage > stage for u in self.users)


def recursive_add_node(cls, name: str, prop: Optional, dims: Shape, nodes: Dict[str, PGraphNode]) -> PGraphNode:
    if name in nodes:
        return nodes[name]
    prop = getattr(cls, name) if prop is None else prop
    requires = get_required_dims(prop, dims)
    node = nodes[name] = PGraphNode(name, prop, frozenset(requires.names))
    dep_names = get_dependencies(cls, prop)
    node.dependencies = [recursive_add_node(cls, n, None, dims, nodes) for n in dep_names]
    for dep in node.dependencies:
        dep.users.append(node)
    return node


def build_stages(nodes: Dict[str, PGraphNode]) -> List[List[PGraphNode]]:
    """ Groups nodes by same `requires`, taking dependencies into account. """
    stages = []
    while any(not n.done for n in nodes.values()):
        candidates = [n for n in nodes.values() if n.can_run_now]
        candidate_req = set([cn.requires for cn in candidates])
        stage_req = next(iter(candidate_req))
        stage = []
        for n in candidates:
            if n.requires == stage_req:
                n.done = True
                n.stage = len(stages)
                stage.append(n)
        stages.append(stage)
    return stages


def parallel_property(func: Callable = None, /, requires: DimFilter = None, on_direct_eval='raise'):
    """
    Similar to `@cached_property` but with additional controls over parallelization.

    See Also:
        `parallel_compute()`.

    Args:
        func: Method to wrap.
        requires: Dimensions which must be present within one process. These cannot be parallelized when computing this property.
        on_direct_eval: What to do when the property is accessed normally (outside `parallel_compute`) before it has been computed.
            Option:

            * `'raise'`: Raise an error.
            * `'host-compute'`: Compute the property directly, without using multi-threading.
    """
    assert on_direct_eval in {'host-compute', 'raise'}
    if func is None:
        return partial(parallel_property, requires=requires, on_direct_eval=on_direct_eval)
    return ParallelProperty(func, requires=requires, on_direct_eval=on_direct_eval)


_NOT_CACHED = object()


class ParallelProperty(cached_property):
    def __init__(self, func, requires: DimFilter, on_direct_eval: str):
        super().__init__(func)
        self.requires = requires
        self.on_direct_eval = on_direct_eval

    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self.attrname is None:
            raise TypeError("Cannot use cached_property instance without calling __set_name__ on it.")
        try:
            cache = instance.__dict__
        except AttributeError:  # not all objects have __dict__ (e.g. class defines slots)
            raise TypeError(f"No '__dict__' attribute on {type(instance).__name__!r} instance to cache {self.attrname!r} property.") from None
        val = cache.get(self.attrname, _NOT_CACHED)
        if val is _NOT_CACHED:
            if self.on_direct_eval == 'host-compute':
                val = super().__get__(instance, owner=owner)
            elif self.on_direct_eval == 'raise':
                raise NonParallelAccess(f"@parallel_property '{self.attrname}' can only be accessed after evaluation by parallel_compute()")
        return val

    def get_on_host(self, instance):
        return super().__get__(instance, owner=None)

    # __class_getitem__ = classmethod(GenericAlias)  # inherited from cached_property


class NonParallelAccess(Exception):
    pass


def get_property_value(instance, name: str):
    prop = getattr(type(instance), name)
    if isinstance(prop, ParallelProperty):
        return prop.get_on_host(instance)
    else:
        return prop.__get__(instance)


def property_name(p):
    if isinstance(p, property):
        return p.fget.__name__
    elif isinstance(p, cached_property):
        return p.func.__name__
    elif hasattr(p, '__qualname__'):
        return p.__qualname__
    else:
        raise ValueError(f"Not a property: {p}")


def get_required_dims(p, dims: Shape) -> Shape:
    if isinstance(p, ParallelProperty):
        return dims.only(p.requires)
    return EMPTY_SHAPE



from dataclasses import dataclass
from functools import lru_cache
from typing import Union

import numpy as np
import sys


class DTypeMeta(type):

    def __call__(self, *args, **kwargs):
        if len(args) == 1 and not kwargs:
            return {bool: BOOL, object: OBJECT}[args[0]]
        if len(args) == 2 and not kwargs:
            return DType.by_bits(*args)
        elif 'precision' in kwargs:
            assert len(args) == 1, "If 'precision' is specified, only one argument may be given."
            return DType.by_precision(args[0], kwargs['precision'])
        return type.__call__(self, *args, **kwargs)


@dataclass(frozen=True)
class DType(metaclass=DTypeMeta):
    """
    Instances of `DType` represent the kind and size of data elements.
    The data type of tensors can be obtained via `Tensor.dtype`.

    The following kinds of data types are supported:

    * `float` with 32 / 64 bits
    * `complex` with 64 / 128 bits
    * `int` with 8 / 16 / 32 / 64 bits
    * `bool` with 8 bits
    * `str` with 8*n* bits

    Unlike with many computing libraries, there are no global variables corresponding to the available types.
    Instead, data types can simply be instantiated as needed.
    """
    kind: type
    """Python type, one of `(bool, int, float, complex, str, object)`"""
    bits: int
    """Number of bits per element, typically a multiple of 8."""
    unsigned: bool
    """If `True`, the data type is unsigned, meaning it can only represent non-negative values."""
    exponent_bits: int
    """Number of bits used for the exponent in floating point types. 0 for integers."""
    mantissa_bits: int
    """Number of bits used for the mantissa in floating point types. Same as `bits` for integers."""
    finite_only: bool
    """If `True`, the data type can only represent finite values, i.e., no NaN or Inf."""
    unsigned_zero: bool
    """If `True`, the data type cannot represent signed zeros. This is `True` for integers and `False` for most floating point types."""

    @property
    def precision(self):
        """ Floating point precision. Only defined if `kind in (float, complex)`. For complex values, returns half of `DType.bits`. """
        if self.kind == float:
            return self.bits
        if self.kind == complex:
            return self.bits // 2
        else:
            return None

    @property
    def itemsize(self):
        """ Number of bytes used to storea single value of this type. See `DType.bits`. """
        return self.bits // 8 if self.bits % 8 == 0 else self.bits / 8

    def __eq__(self, other):
        if isinstance(other, DType):
            return self.kind == other.kind and self.bits == other.bits and self.unsigned == other.unsigned
        elif other in {bool, int, float, complex, object}:
            return self.kind == other
        else:
            return False

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.kind)

    def __repr__(self):
        if self.kind == int:
            if self.unsigned:
                return f"uint{self.bits}"
            else:
                return f"int{self.bits}"
        elif self.kind == float:
            if self == FLOAT16:
                return "float16"
            elif self == FLOAT32:
                return "float32"
            elif self == FLOAT64:
                return "float64"
            else:
                return f"float{self.bits}_e{self.exponent_bits}_m{self.mantissa_bits}{'fn' if self.finite_only else ''}{'uz' if self.unsigned_zero else ''}"
        elif self.kind == complex:
            if self == COMPLEX64:
                return "complex64"
            elif self == COMPLEX128:
                return "complex128"
            else:
                return f"complex{self.bits}_e{self.exponent_bits}_m{self.mantissa_bits}{'fn' if self.finite_only else ''}{'uz' if self.unsigned_zero else ''}"
        elif self.kind == str:
            return f"str{self.bits}"
        elif self.kind == bool:
            return "bool"
        elif self.kind == object:
            return "object"
        else:
            return f"{self.kind.__name__}{self.bits}"

    @staticmethod
    def as_dtype(value: Union['DType', tuple, type, None]) -> Union['DType', None]:
        if isinstance(value, DType):
            return value
        elif value is int:
            return INT32
        elif value is float:
            from . import get_precision
            return DType.by_precision(float, get_precision())
        elif value is complex:
            from . import get_precision
            return DType.by_precision(complex, get_precision())
        elif value is None:
            return None
        elif isinstance(value, tuple):
            if len(value) == 2:
                return {
                    (int, 8): INT8,
                    (int, 16): INT16,
                    (int, 32): INT32,
                    (int, 64): INT64,
                    (float, 16): FLOAT16,
                    (float, 32): FLOAT32,
                    (float, 64): FLOAT64,
                    (complex, 64): COMPLEX64,
                    (complex, 128): COMPLEX128,
                }[value]
            return DType(*value)
        elif value is str:
            raise ValueError("str DTypes must specify bits")
        return {bool: BOOL, object: OBJECT}[value]

    @staticmethod
    def by_precision(kind: type, precision: int) -> 'DType':
        if kind == float:
            return {16: FLOAT16, 32: FLOAT32, 64: FLOAT64}[precision]
        elif kind == complex:
            return {32: COMPLEX64, 64: COMPLEX128}[precision]
        else:
            raise ValueError(f"Unsupported kind: {kind}")

    @staticmethod
    def int_by_bits(bits: int):
        return {8: INT8, 16: INT16, 32: INT32, 64: INT64}[bits]

    @staticmethod
    def by_bits(kind: type, bits: int):
        if kind is int:
            return {8: INT8, 16: INT16, 32: INT32, 64: INT64}[bits]
        elif kind is float:
            return {16: FLOAT16, 32: FLOAT32, 64: FLOAT64}[bits]
        elif kind is complex:
            return {64: COMPLEX64, 128: COMPLEX128}[bits]
        elif kind is str:
            return DType(str, bits, False, 0, 0, True, True)
        raise ValueError


# def get_dtype(kind: type, bits: int = None, precision: int = None):
#     assert kind in (bool, int, float, complex, str, object)
#     if kind is bool:
#         assert precision is None, f"Precision may only be specified for float or complex but got {kind}, precision={precision}"
#         return BOOL if bits is None else DType(bool, bits=bits, unsigned=True, exponent_bits=0, mantissa_bits=1, finite_only=True, unsigned_zero=True)
#     elif kind == object:
#         assert precision is None, f"Precision may only be specified for float or complex but got {kind}, precision={precision}"
#         return OBJECT if bits is None else DType(object, bits=bits, unsigned=True, exponent_bits=0, mantissa_bits=bits, finite_only=False, unsigned_zero=False)
#     elif precision is not None:
#         assert bits is None, "Specify either bits or precision when creating a DType but not both."
#         assert kind in [float, complex], f"Precision may only be specified for float or complex but got {kind}, precision={precision}"
#         if kind == float:
#             bits = precision
#         else:
#             bits = precision * 2
#     else:
#         assert isinstance(bits, int), f"bits must be an int but got {type(bits)}"


# --- NumPy Conversion ---

def to_numpy_dtype(dtype: DType):
    if dtype in _TO_NUMPY:
        return _TO_NUMPY[dtype]
    if dtype.kind == str:
        bytes_per_char = np.dtype('<U1').itemsize
        return np.dtype(f'<U{dtype.itemsize // bytes_per_char}')
    raise KeyError(f"Unsupported dtype: {dtype}")


def from_numpy_dtype(np_dtype) -> DType:
    if np_dtype in _FROM_NUMPY:
        return _FROM_NUMPY[np_dtype]
    else:
        for base_np_dtype, dtype in _FROM_NUMPY.items():
            if np_dtype == base_np_dtype:
                _FROM_NUMPY[np_dtype] = dtype
                return dtype
        if np_dtype.char == 'U':
            return DType(str, 8 * np_dtype.itemsize)
        raise ValueError(np_dtype)


BOOL = DType(bool, bits=8, unsigned=True, exponent_bits=0, mantissa_bits=1, finite_only=True, unsigned_zero=True)
OBJECT = DType(object, bits=int(np.round(np.log2(sys.maxsize))) + 1, unsigned=True, exponent_bits=0, mantissa_bits=int(np.round(np.log2(sys.maxsize))) + 1, finite_only=False, unsigned_zero=False)
# --- Int ---
INT8 = DType(int, 8, unsigned=False, exponent_bits=0, mantissa_bits=8, finite_only=True, unsigned_zero=True)
INT16 = DType(int, 16, unsigned=False, exponent_bits=0, mantissa_bits=16, finite_only=True, unsigned_zero=True)
INT32 = DType(int, 32, unsigned=False, exponent_bits=0, mantissa_bits=32, finite_only=True, unsigned_zero=True)
INT64 = DType(int, 64, unsigned=False, exponent_bits=0, mantissa_bits=64, finite_only=True, unsigned_zero=True)
UINT8 = DType(int, 8, unsigned=True, exponent_bits=0, mantissa_bits=8, finite_only=True, unsigned_zero=True)
UINT16 = DType(int, 16, unsigned=True, exponent_bits=0, mantissa_bits=16, finite_only=True, unsigned_zero=True)
UINT32 = DType(int, 32, unsigned=True, exponent_bits=0, mantissa_bits=32, finite_only=True, unsigned_zero=True)
UINT64 = DType(int, 64, unsigned=True, exponent_bits=0, mantissa_bits=64, finite_only=True, unsigned_zero=True)
# --- Float ---
BF16 = DType(float, 16, unsigned=False, exponent_bits=8, mantissa_bits=7, finite_only=False, unsigned_zero=False)
FLOAT16 = DType(float, 16, unsigned=False, exponent_bits=5, mantissa_bits=10, finite_only=False, unsigned_zero=False)
FLOAT32 = DType(float, 32, unsigned=False, exponent_bits=8, mantissa_bits=23, finite_only=False, unsigned_zero=False)
FLOAT64 = DType(float, 64, unsigned=False, exponent_bits=11, mantissa_bits=52, finite_only=False, unsigned_zero=False)
# --- Complex ---
COMPLEX64 = DType(complex, 64, unsigned=False, exponent_bits=8, mantissa_bits=23, finite_only=False, unsigned_zero=False)
COMPLEX128 = DType(complex, 128, unsigned=False, exponent_bits=11, mantissa_bits=52, finite_only=False, unsigned_zero=False)

_TO_NUMPY = {
    BOOL: np.bool_,
    OBJECT: object,
    # --- Int ---
    INT8: np.int8,
    INT16: np.int16,
    INT32: np.int32,
    INT64: np.int64,
    UINT8: np.uint8,
    UINT16: np.uint16,
    UINT32: np.uint32,
    UINT64: np.uint64,
    # --- Float ---
    FLOAT16: np.float16,
    FLOAT32: np.float32,
    FLOAT64: np.float64,
    # --- Complex ---
    COMPLEX64: np.complex64,
    COMPLEX128: np.complex128,
}
_FROM_NUMPY = {np: dtype for dtype, np in _TO_NUMPY.items()}
_FROM_NUMPY[np.bool_] = BOOL
_FROM_NUMPY[bool] = BOOL


UPCAST_LVL = {bool: 1, int: 2, float: 3, complex: 4, str: 5, object: 6}


@lru_cache
def combine_types(*dtypes: DType, fp_precision: int = None) -> DType:
    lvl = max(UPCAST_LVL[dt.kind] for dt in dtypes)
    if lvl == 1:  # all bool
        return dtypes[0]
    elif lvl == 2:  # all int / bool?
        return max(dtypes, key=lambda dt: dt.mantissa_bits)
    elif lvl == 3:  # all real?
        if isinstance(fp_precision, int):
            return DType.by_precision(float, fp_precision)
        else:
            highest_fp = max(dt.precision for dt in dtypes if dt.kind == float)
            return DType.by_precision(float, highest_fp)
    elif lvl == 4:  # complex
        if isinstance(fp_precision, int):
            return DType.by_precision(complex, fp_precision)
        else:
            highest_fp = max([dt.precision for dt in dtypes if dt.kind in (float, complex)])
            return DType.by_precision(complex, highest_fp)
    elif lvl == 5:  # string
        return max([dt for dt in dtypes if dt.kind == str], key=lambda dt: dt.bits)
    return OBJECT



"""
Unified neural network library.
Includes

* Flexible NN creation of popular architectures
* Optimizer creation
* Training functionality
* Parameter access
* Saving and loading networks and optimizer states.
"""
import numpy as np
import os
import time
import warnings
from dataclasses import dataclass
from itertools import count
from typing import Callable, Union, Sequence, Dict, TypeVar, Any, Optional

from .backend import default_backend, Backend, BACKENDS, ComputeDevice
from .backend._backend import init_backend
from . import math, non_channel
from .math import Tensor, Shape, use as _use, batch, layout, load, shape, convert, to_device, pack_dims, EMPTY_SHAPE, DimFilter
from .math.perm import random_permutation

use = _use


def _native_lib():
    if default_backend().supports(Backend.nn_library):
        return default_backend().nn_library()
    ml_backends = [b for b in BACKENDS if b.supports(Backend.nn_library)]
    if not ml_backends:
        if not init_backend('all-imported'):
            raise RuntimeError(f"No ML library available. Please import either jax, torch, or tensorflow.")
        ml_backends = [b for b in BACKENDS if b.supports(Backend.nn_library)]
    if len(ml_backends) > 1:
        warnings.warn(f"Multiple ML libraries loaded {tuple(ml_backends)} but none set as default. Defaulting to {ml_backends[0]}.", RuntimeWarning, stacklevel=3)
    return ml_backends[0].nn_library()


Network = TypeVar('Network')
Optimizer = TypeVar('Optimizer')


def parameter_count(net: Network) -> int:
    """
    Counts the number of parameters in a model.

    See Also:
        `get_parameters()`.

    Args:
        net: PyTorch model

    Returns:
        Total parameter count as `int`.
    """
    return sum([value.shape.volume for name, value in get_parameters(net).items()])


def get_parameters(net: Network) -> Dict[str, Tensor]:
    """
    Returns all parameters of a neural network.

    Args:
        net: Neural network.

    Returns:
        `dict` mapping parameter names to `phiml.math.Tensor`s.
    """
    return _native_lib().get_parameters(net)


def save_state(obj: Union[Network, Optimizer], path: str) -> str:
    """
    Write the state of a module or optimizer to a file.

    See Also:
        `load_state()`

    Args:
        obj: `torch.Network or torch.optim.Optimizer`
        path: File path as `str`.

    Returns:
        Path to the saved file.
    """
    return _native_lib().save_state(**locals())


def load_state(obj: Union[Network, Optimizer], path: str):
    """
    Read the state of a module or optimizer from a file.

    See Also:
        `save_state()`

    Args:
        obj: `torch.Network or torch.optim.Optimizer`
        path: File path as `str`.
    """
    return _native_lib().load_state(**locals())


def update_weights(net: Network, optimizer: Optimizer, loss_function: Callable, *loss_args, **loss_kwargs):
    """
    Computes the gradients of `loss_function` w.r.t. the parameters of `net` and updates its weights using `optimizer`.

    This is the PyTorch version. Analogue functions exist for other learning frameworks.

    Args:
        net: Learning model.
        optimizer: Optimizer.
        loss_function: Loss function, called as `loss_function(*loss_args, **loss_kwargs)`.
        *loss_args: Arguments given to `loss_function`.
        **loss_kwargs: Keyword arguments given to `loss_function`.

    Returns:
        Output of `loss_function`.
    """
    return _native_lib().update_weights(net, optimizer, loss_function, *loss_args, **loss_kwargs)


def adam(net: Network, learning_rate: float = 1e-3, betas=(0.9, 0.999), epsilon=1e-07):
    """
    Creates an Adam optimizer for `net`, alias for [`torch.optim.Adam`](https://pytorch.org/docs/stable/generated/torch.optim.Adam.html).
    Analogue functions exist for other learning frameworks.
    """
    return _native_lib().adam(**locals())


def sgd(net: Network, learning_rate: float = 1e-3, momentum=0., dampening=0., weight_decay=0., nesterov=False):
    """
    Creates an SGD optimizer for 'net', alias for ['torch.optim.SGD'](https://pytorch.org/docs/stable/generated/torch.optim.SGD.html)
    Analogue functions exist for other learning frameworks.
    """
    return _native_lib().sgd(**locals())


def adagrad(net: Network, learning_rate: float = 1e-3, lr_decay=0., weight_decay=0., initial_accumulator_value=0., eps=1e-10):
    """
    Creates an Adagrad optimizer for 'net', alias for ['torch.optim.Adagrad'](https://pytorch.org/docs/stable/generated/torch.optim.Adagrad.html)
    Analogue functions exist for other learning frameworks.
    """
    return _native_lib().adagrad(**locals())


def rmsprop(net: Network, learning_rate: float = 1e-3, alpha=0.99, eps=1e-08, weight_decay=0., momentum=0., centered=False):
    """
    Creates an RMSProp optimizer for 'net', alias for ['torch.optim.RMSprop'](https://pytorch.org/docs/stable/generated/torch.optim.RMSprop.html)
    Analogue functions exist for other learning frameworks.
    """
    return _native_lib().rmsprop(**locals())


def set_learning_rate(optimizer: Optimizer, learning_rate: Union[float, Tensor]):
    """
    Sets the global learning rate for the given optimizer.

    Args:
        optimizer (optim.Optimizer): The optimizer whose learning rate needs to be updated.
        learning_rate (float): The new learning rate to set.
    """
    _native_lib().set_learning_rate(optimizer, float(learning_rate))


def get_learning_rate(optimizer: Optimizer) -> float:
    """
    Returns the global learning rate of the given optimizer.

    Args:
        optimizer (optim.Optimizer): The optimizer whose learning rate needs to be retrieved.

    Returns:
        float: The learning rate of the optimizer.
    """
    return _native_lib().get_learning_rate(optimizer)


def mlp(in_channels: int,
              out_channels: int,
              layers: Sequence[int],
              batch_norm=False,
              activation: Union[str, Callable] = 'ReLU',
              softmax=False) -> Network:
    """
    Fully-connected neural networks are available in Φ-ML via mlp().

    Args:
        in_channels: size of input layer, int
        out_channels = size of output layer, int
        layers: tuple of linear layers between input and output neurons, list or tuple
        activation: activation function used within the layers, string
        batch_norm: use of batch norm after each linear layer, bool

    Returns:
        Dense net model as specified by input arguments
    """
    return _native_lib().mlp(**locals())


def u_net(in_channels: int,
          out_channels: int,
          levels: int = 4,
          filters: Union[int, Sequence] = 16,
          batch_norm: bool = True,
          activation: Union[str, type] = 'ReLU',
          in_spatial: Union[tuple, int] = 2,
          periodic=False,
          use_res_blocks: bool = False,
          down_kernel_size=3,
          up_kernel_size=3) -> Network:
    """
    Built-in U-net architecture, classically popular for Semantic Segmentation in Computer Vision, composed of downsampling and upsampling layers.

    Args:
        in_channels: input channels of the feature map, dtype: int
        out_channels: output channels of the feature map, dtype: int
        levels: number of levels of down-sampling and upsampling, dtype: int
        filters: filter sizes at each down/up sampling convolutional layer, if the input is integer all conv layers have the same filter size,
        activation: activation function used within the layers, dtype: string
        batch_norm: use of batchnorm after each conv layer, dtype: bool
        in_spatial: spatial dimensions of the input feature map, dtype: int
        use_res_blocks: use convolutional blocks with skip connections instead of regular convolutional blocks, dtype: bool
        down_kernel_size: Kernel size for convolutions on the down-sampling (first half) side of the U-Net.
        up_kernel_size: Kernel size for convolutions on the up-sampling (second half) of the U-Net.

    Returns:
        U-net model as specified by input arguments.
    """
    return _native_lib().u_net(**locals())


def conv_net(in_channels: int,
             out_channels: int,
             layers: Sequence[int],
             batch_norm: bool = False,
             activation: Union[str, type] = 'ReLU',
             in_spatial: Union[int, tuple] = 2,
             periodic=False) -> Network:
    """
    Built in Conv-Nets are also provided. Contrary to the classical convolutional neural networks, the feature map spatial size remains the same throughout the layers. Each layer of the network is essentially a convolutional block comprising of two conv layers. A filter size of 3 is used in the convolutional layers.

    Args:
        in_channels: input channels of the feature map, dtype: int
        out_channels: output channels of the feature map, dtype: int
        layers: list or tuple of output channels for each intermediate layer between the input and final output channels, dtype: list or tuple
        activation: activation function used within the layers, dtype: string
        batch_norm: use of batchnorm after each conv layer, dtype: bool
        in_spatial: spatial dimensions of the input feature map, dtype: int

    Returns:
        Conv-net model as specified by input arguments
    """
    return _native_lib().conv_net(**locals())


def res_net(in_channels: int,
            out_channels: int,
            layers: Sequence[int],
            batch_norm: bool = False,
            activation: Union[str, type] = 'ReLU',
            in_spatial: Union[int, tuple] = 2,
            periodic=False) -> Network:
    """
    Similar to the conv-net, the feature map spatial size remains the same throughout the layers.
    These networks use residual blocks composed of two conv layers with a skip connection added from the input to the output feature map.
    A default filter size of 3 is used in the convolutional layers.

    Args:
        in_channels: input channels of the feature map, dtype: int
        out_channels: output channels of the feature map, dtype: int
        layers: list or tuple of output channels for each intermediate layer between the input and final output channels, dtype: list or tuple
        activation: activation function used within the layers, dtype: string
        batch_norm: use of batchnorm after each conv layer, dtype: bool
        in_spatial: spatial dimensions of the input feature map, dtype: int

    Returns:
        Res-net model as specified by input arguments
    """
    return _native_lib().res_net(**locals())


def conv_classifier(in_features: int,
                    in_spatial: Union[tuple, list],
                    num_classes: int,
                    blocks=(64, 128, 256, 256, 512, 512),
                    block_sizes=(2, 2, 3, 3, 3),
                    dense_layers=(4096, 4096, 100),
                    batch_norm=True,
                    activation='ReLU',
                    softmax=True,
                    periodic=False):
    """
    Based on VGG16.
    """
    return _native_lib().conv_classifier(**locals())


def invertible_net(num_blocks: int = 3,
                   construct_net: Union[str, Callable] = 'u_net',
                   **construct_kwargs):
    """
    Invertible NNs are capable of inverting the output tensor back to the input tensor initially passed.
    These networks have far-reaching applications in predicting input parameters of a problem given its observations.
    Invertible nets are composed of multiple concatenated coupling blocks wherein each such block consists of arbitrary neural networks.

    Currently, these arbitrary neural networks could be set to u_net(default), conv_net, res_net or mlp blocks with in_channels = out_channels.
    The architecture used is popularized by ["Real NVP"](https://arxiv.org/abs/1605.08803).

    Invertible nets are only implemented for PyTorch and TensorFlow.

    Args:
        num_blocks: number of coupling blocks inside the invertible net, dtype: int
        construct_net: Function to construct one part of the neural network.
            This network must have the same number of inputs and outputs.
            Can be a `lambda` function or one of the following strings: `mlp, u_net, res_net, conv_net`
        construct_kwargs: Keyword arguments passed to `construct_net`.

    Returns:
        Invertible neural network model
    """
    return _native_lib().invertible_net(num_blocks, construct_net, **construct_kwargs)


# def fno(in_channels: int,
#         out_channels: int,
#         mid_channels: int,
#         modes: Sequence[int],
#         activation: Union[str, type] = 'ReLU',
#         batch_norm: bool = False,
#         in_spatial: int = 2):
#     """
#     ["Fourier Neural Operator"](https://github.com/zongyi-li/fourier_neural_operator) network contains 4 layers of the Fourier layer.
#     1. Lift the input to the desire channel dimension by self.fc0 .
#     2. 4 layers of the integral operators u' = (W + K)(u). W defined by self.w; K defined by self.conv .
#     3. Project from the channel space to the output space by self.fc1 and self.fc2.
#
#     Args:
#         in_channels: input channels of the feature map, dtype: int
#         out_channels: output channels of the feature map, dtype: int
#         mid_channels: channels used in Spectral Convolution Layers, dtype: int
#         modes: Fourier modes for each spatial channel, dtype: List[int] or int (in case all number modes are to be the same for each spatial channel)
#         activation: activation function used within the layers, dtype: string
#         batch_norm: use of batchnorm after each conv layer, dtype: bool
#         in_spatial: spatial dimensions of the input feature map, dtype: int
#
#     Returns:
#         Fourier Neural Operator model as specified by input arguments.
#     """
#     return _native_lib().fno(**locals())


@dataclass(frozen=True)
class TrainingState:
    name: str
    model: Network
    optimizer: Optimizer
    learning_rate: float
    epoch: int
    max_epochs: Optional[int]
    iter: int
    max_iter: Optional[int]
    is_epoch_end: bool
    epoch_loss: Tensor
    batch_loss: Optional[Tensor]
    additional_batch_output: Optional[tuple]
    indices: Tensor
    termination_reason: Optional[str]
    peak_memory: Optional[int]

    @property
    def current(self) -> int:
        return self.epoch if self.is_epoch_end else self.iter

    @property
    def max(self) -> int:
        return self.max_epochs if self.is_epoch_end else self.max_iter

    @property
    def mean_loss(self) -> float:
        return float(self.epoch_loss) if self.is_epoch_end else float(math.mean(self.batch_loss, 'dset_linear'))


def train(name: Optional[str], model, optimizer, loss_fn: Callable,
          *files_or_data: Union[str, Tensor],
          max_epochs: int = None, max_iter: int = None, max_hours: float = None, stop_on_loss: float = None,
          batch_size: int = 1, file_shape: Shape = EMPTY_SHAPE, dataset_dims: DimFilter = batch, device: ComputeDevice = None, drop_last=False, loss_kwargs=None,
          lr_schedule_iter=None, checkpoint_frequency=None, loader=math.load,
          on_iter_end: Callable = None, on_epoch_end: Callable = None,
          measure_peak_memory: bool = True) -> TrainingState:
    """
    Call `update_weights()` for each batch in a loop for each epoch.

    Args:
        name: Name of the model. This is used as a name to save the model and optimizer states.
            If not specified, no model or optimizer states are saved.
        model: PyTorch module or Keras model (TensorFlow).
        optimizer: PyTorch or TensorFlow/Keras optimizer.
        loss_fn: Loss function for training. This function should take the data as input, run the model and return the loss. It may return additional outputs, but the loss must be the first value.
        *files_or_data: Training data or file names containing training data or a mixture of both. Files are loaded using `loader`.
        max_epochs: Epoch limit.
        max_iter: Iteration limit. The number of iterations depends on the batch size and the number of files.
        max_hours: Training time limit in hours (`float`).
        stop_on_loss: Stop training if the mean epoch loss falls below this value.
        batch_size: Batch size for training. The batch size is limited by the number of data points in the dataset.
        file_shape: Shape of data stored in each file.
        dataset_dims: Which dims of the training data list training examples, as opposed to features of data points.
        device: Device to use for training. If `None`, the default device is used.
        drop_last: If `True`, drop the last batch if it is smaller than `batch_size`.
        loss_kwargs: Keyword arguments passed to `loss_fn`.
        lr_schedule_iter: Function `(i: int) -> float` that returns the learning rate for iteration `i`. If `None`, the learning rate of the `optimizer` is used as is.
        checkpoint_frequency: If not `None`, save the model and optimizer state every `checkpoint_frequency` epochs.
        loader: Function `(file: str) -> data: Tensor` to load data from files. Defaults to `phiml.math.load()`.
        on_iter_end: Function `(i: int, max_iter: int, name: str, model, optimizer, learning_rate, loss, *additional_output) -> None` called after each iteration. The function is called with the current iteration number `i` starting at 0, the maximum number of iterations `max_iter`, the name of the model `name`, the model `model`, the optimizer `optimizer`, the learning rate `learning_rate`, the loss value `loss` and any additional output from `loss_fn`.
        on_epoch_end: Function `(epoch: int, max_epochs: int, name: str, model, optimizer, learning_rate, epoch_loss) -> None` called after each epoch. The function is called with the current epoch number `epoch` starting at 0, the maximum number of epochs `max_epochs`, the name of the model `name`, the model `model`, the optimizer `optimizer`, the learning rate `learning_rate` and the average loss for the epoch `epoch_loss`.
        measure_peak_memory: If `True`, measure the peak memory usage during training and store it in the returned `TrainingState`. This is only supported by some backends.

    Returns:
        `TrainingResult` containing the termination reason, last epoch and last iteration.
    """
    files_or_data = [layout(fs) if isinstance(fs, str) else fs for fs in files_or_data]
    data_shape = shape(files_or_data) & file_shape
    loss_kwargs = {} if loss_kwargs is None else loss_kwargs
    device = device if device is not None else default_backend().get_default_device()
    if measure_peak_memory:
        default_backend().reset_peak_memory(device)
    default_backend().set_default_device('CPU')
    data = [convert(math.map(lambda f: loader(f), fs, map_name="Loading data")) if isinstance(fs, str) or (isinstance(fs, Tensor) and fs.dtype.kind == object) else fs for fs in files_or_data]
    default_backend().set_default_device(device)
    data_shape = shape(data)
    dataset_dims = data_shape.only(dataset_dims)
    batch_size = min(batch_size, dataset_dims.volume)
    batch_count = dataset_dims.volume // batch_size if drop_last else (dataset_dims.volume + batch_size - 1) // batch_size
    name and os.makedirs(name, exist_ok=True)
    learning_rate = None if lr_schedule_iter is not None else get_learning_rate(optimizer)
    if max_epochs is None and max_iter is not None:
        max_epochs = int(np.ceil(max_iter / batch_count))
    elif max_epochs is not None and max_iter is not None:
        max_iter = max_epochs * batch_count
    termination_reason = None
    niter = 0
    epoch = 0
    t0 = time.perf_counter()
    for epoch in range(max_epochs) if max_epochs is not None else count():
        default_backend().set_default_device('CPU')
        indices = pack_dims(random_permutation(dataset_dims, dims=dataset_dims), non_channel, batch('dset_linear'))
        default_backend().set_default_device(device)
        epoch_loss = 0
        for i in range(batch_count):
            if lr_schedule_iter is not None:
                learning_rate = lr_schedule_iter(niter)
                set_learning_rate(optimizer, learning_rate)
            batch_idx = indices.dset_linear[i * batch_size:(i + 1) * batch_size]
            data_batch = to_device(math.slice(data, batch_idx), device)
            output = update_weights(model, optimizer, loss_fn, *data_batch, **loss_kwargs)
            loss, *additional_output = output if isinstance(output, (tuple, list)) else (output,)
            epoch_loss += math.sum(loss, 'dset_linear')
            if on_iter_end is not None:
                try:
                    on_iter_end(TrainingState(name, model, optimizer, learning_rate, epoch, max_epochs, niter, max_iter, False, epoch_loss / ((i+1)*batch_count), loss, additional_output, batch_idx, None, None))
                except StopTraining as stop:
                    termination_reason = stop.reason
                    break
            niter += 1
            if max_iter is not None and niter >= max_iter:
                termination_reason = 'max_iter'
                break
            if max_hours is not None and time.perf_counter() - t0 > max_hours * 3600:
                termination_reason = 'max_hours'
                break
        if termination_reason is not None:
            break
        if name and checkpoint_frequency is not None and (epoch + 1) % checkpoint_frequency == 0:
            save_state(model, f"{name}/model_{epoch + 1}")
            save_state(optimizer, f"{name}/optimizer_{epoch + 1}")
        epoch_loss /= indices.dset_linear.size
        if on_epoch_end is not None:
            try:
                on_epoch_end(TrainingState(name, model, optimizer, learning_rate, epoch, max_epochs, niter, max_iter, True, epoch_loss, None, None, indices, None, None))
            except StopTraining as stop:
                termination_reason = stop.reason
                break
        if stop_on_loss is not None and (epoch_loss.mean < stop_on_loss):
            termination_reason = 'stop_on_loss'
            break
    peak_memory = default_backend().get_peak_memory(device) if measure_peak_memory else None
    return TrainingState(name, model, optimizer, learning_rate, epoch, max_epochs, niter, max_iter, True, None, None, None, None, termination_reason, peak_memory)


class StopTraining(Exception):
    """ This exception is raised by the `on_epoch_end` or `on_iter_end` callbacks to stop training. """
    def __init__(self, reason: str = 'stop'):
        super().__init__(reason)
        self.reason = reason



from typing import TypeVar, Any, Tuple

from ._magic_ops import tree_map, variable_attributes
from ._tensors import Tensor, wrap
from ._ops import convert
from ..backend import default_backend

Obj = TypeVar('Obj')

def module_from_variables(obj: Obj, backend=None) -> Tuple[Obj, Any]:
    backend = backend if backend is not None else default_backend()
    variables = []
    def leaf_param(x: Tensor):
        x = convert(x, backend)
        var = backend.variable(x.native(x.shape))
        variables.append(var)
        return wrap(var, x.shape)
    tree = tree_map(leaf_param, obj, attr_type=variable_attributes)
    module = backend.module({f'param_{i}': p for i, p in enumerate(variables)})
    return tree, module



# This module is only imported on demand
from typing import Callable

import torch

from .._backend import get_backend

from ... import Tensor, tensor, batch, channel, spatial, math

TORCH = get_backend('torch')


def handle_torch_function(func: Callable, types, args: tuple, kwargs: dict):  # called by Tensor.__torch_function__
    """
    1. Converts PhiML tensors to PyTorch tensors, even if their native is of another backend.
    2. Calls the PyTorch function.
    3. Converts the result back to PhiML Tensor backed by PyTorch.

    Returns:
        NotImplemented if the function is not supported
    """
    if func == torch.nn.functional.interpolate:
        input: Tensor = args[0]
        if input.backend.name != 'torch':
            input = math.convert(input, TORCH)
        if kwargs['mode'] == 'linear' and spatial(input).rank > 1:
            kwargs['mode'] = 'bilinear' if spatial(input).rank == 2 else 'trilinear'
        native_result = func(input.native([batch, channel, *spatial(input)]), **kwargs)
        return tensor(native_result, [batch(input), channel(input), *spatial(input).without_sizes()], convert=False)
    return NotImplemented
